import type { Logger } from 'winston';
import net from 'net';
import type { Socket } from 'net';
import { createLog } from '../log'
import { Shell } from '../shell';
import { delay } from '../delay';

enum Action {
	NONE,
	LOAD,
	DISPLAY,
	STOP
}

enum Mode {
	RGB,
	BW,
	INVERT,
	BW_INVERT,
	RGB_CHANNELS,
	INVERT_CHANNELS
}

interface fdOutgoingPosition {
	x : number,
	y : number,
	w : number,
	h : number
}

interface fdOutgoingMessage {
	action : Action,
	image : string,
	mode? : Mode,
	position? : fdOutgoingPosition,
	exposure? : number[]
}

interface fdIncomingMessage {
	action : Action,
	success : boolean,
	error? : string
}

interface fdResult {
	action : Action,
	image : string,
	time : number
}

export class FD {
	private bin : string;
	private width : number;
	private height : number;
	private host : string;
	private port : number;
	private shell : Shell;
	private log : Logger;
	private client : Socket;
	private socketAvailable : boolean = false;
	private socketConnected : boolean = false;

	private waiting : Function = null;
	private mock : boolean = false;

	constructor (bin: string, width : number, height : number, host : string, port : number, mock : boolean = false) {
		this.bin = bin;
		this.width = width;
		this.height = height;
		this.host = host;
		this.port = port;
		this.mock = mock;
		this.log = createLog('fd');
		if (!this.mock) this.shell = new Shell([ this.bin, `${this.width}`, `${this.height}`, `${this.port}` ], this.logstd.bind(this), this.logsterr.bind(this), null, true);
		this.startDisplay();
		this.startClient();
		//this.test();
	}

	private async startDisplay () {
		this.log.info(`Launching fd binary ${this.bin}`);
		if (!this.mock) this.shell.execute();
	}

	private async startClient () {
		if (this.mock) {
			return false;
		}
		this.client = new net.Socket();
		this.log.info(`Waiting for TCP socket server on ${this.host}:${this.port}...`);
		while (!this.socketAvailable) {
			await delay(10);
		}
		this.log.info(`TCP socket server on ${this.host}:${this.port} is available`);
		this.client.connect(this.port, this.host, async function () {
			this.socketConnected = true;
			this.log.info(`TCP socket client connected to ${this.host}:${this.port}`);
		}.bind(this));

		this.client.on('data', function (data : string) {
		    this.receive(data);
		}.bind(this));

		this.client.on('close', () => {
		    this.log.info('Closing connection');
		    this.socketConnected = false;
		});

		this.client.on('error', (err : Error) => {
		    this.log.error('Error in socket client', err);
		    this.socketConnected = false;
		    if (!this.mock) this.shell.kill();
		});
	}

	private logstd (data : string ) {
		let parsed : any = null;
		try {
			parsed = JSON.parse(data);
		} catch (err) {
			this.log.error(`Error parsing line ${data}`, err);
		}
		if (parsed !== null) {
			this.log.info(`[shell] ${data}`);
			if (typeof parsed['listening'] !== 'undefined' && parsed['listening'] === true
				&& typeof parsed['port'] !== 'undefined' && parsed['port'] === this.port) {
				this.socketAvailable = true;
			}
		}
	}

	private logsterr (data : string) {
		this.log.error(`[shell] ${data}`);
	}

	private send (msg : fdOutgoingMessage) {
		const json : string = JSON.stringify(msg);
		this.log.info(json);
		if (!this.mock) this.client.write(json);
	}

	private receive (json : string) {
		const msg : fdIncomingMessage = JSON.parse(json);
		this.log.info(msg);
		if (this.waiting != null) {
			this.waiting(msg);
			this.waiting = null;
		}
	}

	public async load (image : string, x : number, y : number, w : number, h : number) : Promise<fdResult> {
		const msg : fdOutgoingMessage = {
			action : Action.LOAD,
			image,
			position : {
				x,
				y,
				w,
				h
			}
		};
		const startTime : number = +new Date();
		if (this.mock) {
			return {
				action : Action.LOAD,
				image,
				time : (+new Date()) - startTime
			};
		}
		const promise : Promise<fdResult> = new Promise(function (resolve : Function, reject : Function) {
			this.waiting = function (msg : fdIncomingMessage) {
				if (msg.action == Action.LOAD && msg.success) {
					return resolve({
						action : Action.LOAD,
						image,
						time : (+new Date()) - startTime
					});
				} else if (typeof msg.error !== 'undefined') {
					return reject(new Error(msg.error));
				}
				return reject(new Error(`Error loading ${image}`));
			}
		}.bind(this));
		this.send(msg);
		return await promise;
	}

	public async display (image : string, exposure : number[] = []) : Promise<fdResult> {
		const msg : fdOutgoingMessage = {
			action : Action.DISPLAY,
			image,
			exposure
		};
		const startTime : number = +new Date();
		if (this.mock) {
			for (let exp of exposure) {
				await delay(exp);
			}
			return {
				action : Action.DISPLAY,
				image,
				time : (+new Date()) - startTime
			};
		}
		const promise : Promise<fdResult> = new Promise(function (resolve : Function, reject : Function) {
			this.waiting = function (msg : fdIncomingMessage) {
				if (msg.action == Action.DISPLAY && msg.success) {
					return resolve({
						action : Action.DISPLAY,
						image,
						time : (+new Date()) - startTime
					});
				} else if (typeof msg.error !== 'undefined') {
					return reject(new Error(msg.error));
				}
				return reject(new Error(`Error displaying ${image}`));
			}
		}.bind(this));
		this.send(msg);
		return await promise;
	}

	public async stop (image : string) : Promise<fdResult> {
		const msg : fdOutgoingMessage = {
			action : Action.STOP,
			image
		};
		const startTime : number = +new Date();
		if (this.mock) {
			return {
				action : Action.STOP,
				image,
				time : (+new Date()) - startTime
			};
		}
		const promise : Promise<fdResult> = new Promise(function (resolve : Function, reject : Function) {
			this.waiting = function (msg : fdIncomingMessage) {
				if (msg.action == Action.STOP && msg.success) {
					return resolve({
						action : Action.STOP,
						image,
						time : (+new Date()) - startTime
					});
				} else if (typeof msg.error !== 'undefined') {
					return reject(new Error(msg.error));
				}
				return reject(new Error(`Error loading ${image}`));
			}
		}.bind(this));
		this.send(msg);
		return await promise;
	}

	public isConnected () : boolean {
		return this.socketConnected;
	}

	private async test () {
		const img = '/Users/matthewmcwilliams9/src/filmout_display/img/4kSnake.png';
		
		await delay(2000);

		await this.load(img, 100, 200, 640, 640);
		await delay(2000);

		await this.display(img, [ 4000 ] );
		await delay(8000);

		await this.load(img, 100, 200, 640, 640);
		await delay(1000);

		await this.display (img);
		await delay(2000);
		await this.stop(img);

		this.log.warn('QUITTING!!!!');
		process.exit();
	}
}

module.exports = { FD };

export type { Action, Mode, fdOutgoingPosition, fdOutgoingMessage, fdIncomingMessage, fdResult };
